{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "# For tips on running notebooks in Google Colab, see\n# https://pytorch.org/tutorials/beginner/colab\n%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n[\u7b80\u4ecb](introyt1_tutorial.html) ||\n[\u5f20\u91cf](tensors_deeper_tutorial.html) ||\n[\u81ea\u52a8\u5fae\u5206](autogradyt_tutorial.html) ||\n**\u6784\u5efa\u6a21\u578b** ||\n[TensorBoard\u652f\u6301](tensorboardyt_tutorial.html) ||\n[\u8bad\u7ec3\u6a21\u578b](trainingyt.html) ||\n[\u6a21\u578b\u7406\u89e3](captumyt.html)\n\n# \u4f7f\u7528 PyTorch \u6784\u5efa\u6a21\u578b\n\n\u8ddf\u968f\u4e0b\u9762\u7684\u89c6\u9891\u6216\u5728 [youtube](https://www.youtube.com/watch?v=OSqIP-mOWOI)_ \u4e0a\u89c2\u770b.\n\n.. raw:: html\n\n   <div style=\"margin-top:10px; margin-bottom:10px;\">\n     <iframe width=\"560\" height=\"315\" src=\"https://www.youtube.com/embed/OSqIP-mOWOI\" frameborder=\"0\" allow=\"accelerometer; encrypted-media; gyroscope; picture-in-picture\" allowfullscreen></iframe>\n   </div>\n\n## ``torch.nn.Module`` \u548c ``torch.nn.Parameter``\n\n\u5728\u8fd9\u4e2a\u89c6\u9891\u4e2d,\u6211\u4eec\u5c06\u8ba8\u8bba PyTorch \u63d0\u4f9b\u7684\u4e00\u4e9b\u7528\u4e8e\u6784\u5efa\u6df1\u5ea6\u5b66\u4e60\u7f51\u7edc\u7684\u5de5\u5177\u3002\n\n\u9664\u4e86 ``Parameter`` \u4e4b\u5916,\u6211\u4eec\u5728\u672c\u89c6\u9891\u4e2d\u8ba8\u8bba\u7684\u6240\u6709\u7c7b\u90fd\u662f ``torch.nn.Module`` \u7684\u5b50\u7c7b\u3002\n\u8fd9\u662f PyTorch \u7684\u57fa\u7c7b,\u65e8\u5728\u5c01\u88c5\u7279\u5b9a\u4e8e PyTorch \u6a21\u578b\u53ca\u5176\u7ec4\u4ef6\u7684\u884c\u4e3a\u3002\n\n``torch.nn.Module`` \u7684\u4e00\u4e2a\u91cd\u8981\u884c\u4e3a\u662f\u6ce8\u518c\u53c2\u6570\u3002\u5982\u679c\u7279\u5b9a\u7684 ``Module`` \u5b50\u7c7b\u5177\u6709\u5b66\u4e60\u6743\u91cd,\n\u8fd9\u4e9b\u6743\u91cd\u5c06\u8868\u793a\u4e3a ``torch.nn.Parameter`` \u7684\u5b9e\u4f8b\u3002``Parameter`` \u7c7b\u662f ``torch.Tensor`` \u7684\u5b50\u7c7b,\n\u5177\u6709\u7279\u6b8a\u884c\u4e3a,\u5373\u5f53\u5b83\u4eec\u88ab\u5206\u914d\u4e3a ``Module`` \u7684\u5c5e\u6027\u65f6,\u5b83\u4eec\u5c06\u88ab\u6dfb\u52a0\u5230\u8be5\u6a21\u5757\u7684\u53c2\u6570\u5217\u8868\u4e2d\u3002\n\u53ef\u4ee5\u901a\u8fc7 ``Module`` \u7c7b\u4e0a\u7684 ``parameters()`` \u65b9\u6cd5\u8bbf\u95ee\u8fd9\u4e9b\u53c2\u6570\u3002\n\n\u4f5c\u4e3a\u4e00\u4e2a\u7b80\u5355\u7684\u4f8b\u5b50,\u8fd9\u91cc\u6709\u4e00\u4e2a\u975e\u5e38\u7b80\u5355\u7684\u6a21\u578b,\u5305\u542b\u4e24\u4e2a\u7ebf\u6027\u5c42\u548c\u4e00\u4e2a\u6fc0\u6d3b\u51fd\u6570\u3002\n\u6211\u4eec\u5c06\u521b\u5efa\u5b83\u7684\u4e00\u4e2a\u5b9e\u4f8b,\u5e76\u8981\u6c42\u5b83\u62a5\u544a\u5176\u53c2\u6570:\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import torch\n\nclass TinyModel(torch.nn.Module):\n    \n    def __init__(self):\n        super(TinyModel, self).__init__()\n        \n        self.linear1 = torch.nn.Linear(100, 200)\n        self.activation = torch.nn.ReLU()\n        self.linear2 = torch.nn.Linear(200, 10)\n        self.softmax = torch.nn.Softmax()\n    \n    def forward(self, x):\n        x = self.linear1(x)\n        x = self.activation(x)\n        x = self.linear2(x)\n        x = self.softmax(x)\n        return x\n\ntinymodel = TinyModel()\n\nprint('The model:')\nprint(tinymodel)\n\nprint('\\n\\nJust one layer:')\nprint(tinymodel.linear2)\n\nprint('\\n\\nModel params:')\nfor param in tinymodel.parameters():\n    print(param)\n\nprint('\\n\\nLayer params:')\nfor param in tinymodel.linear2.parameters():\n    print(param)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\u8fd9\u663e\u793a\u4e86 PyTorch \u6a21\u578b\u7684\u57fa\u672c\u7ed3\u6784:\u6709\u4e00\u4e2a ``__init__()`` \u65b9\u6cd5\u5b9a\u4e49\u6a21\u578b\u7684\u5c42\u548c\u5176\u4ed6\u7ec4\u4ef6,\n\u8fd8\u6709\u4e00\u4e2a ``forward()`` \u65b9\u6cd5\u6267\u884c\u8ba1\u7b97\u3002\u6ce8\u610f\u6211\u4eec\u53ef\u4ee5\u6253\u5370\u6a21\u578b\u6216\u4efb\u4f55\u5b50\u6a21\u5757,\u4ee5\u4e86\u89e3\u5176\u7ed3\u6784\u3002\n\n## \u5e38\u89c1\u5c42\u7c7b\u578b\n\n### \u7ebf\u6027\u5c42\n\n\u6700\u57fa\u672c\u7684\u795e\u7ecf\u7f51\u7edc\u5c42\u7c7b\u578b\u662f *\u7ebf\u6027* \u6216 *\u5168\u8fde\u63a5* \u5c42\u3002\u8fd9\u662f\u4e00\u79cd\u6bcf\u4e2a\u8f93\u5165\u90fd\u4f1a\u5f71\u54cd\u8be5\u5c42\u6bcf\u4e2a\u8f93\u51fa\u7684\u5c42,\n\u5176\u5f71\u54cd\u7a0b\u5ea6\u7531\u5c42\u7684\u6743\u91cd\u6307\u5b9a\u3002\u5982\u679c\u4e00\u4e2a\u6a21\u578b\u6709 *m* \u4e2a\u8f93\u5165\u548c *n* \u4e2a\u8f93\u51fa,\n\u6743\u91cd\u5c06\u662f\u4e00\u4e2a *m* x *n* \u77e9\u9635\u3002\u4f8b\u5982:\n\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "lin = torch.nn.Linear(3, 2)\nx = torch.rand(1, 3)\nprint('Input:')\nprint(x)\n\nprint('\\n\\nWeight and Bias parameters:')\nfor param in lin.parameters():\n    print(param)\n\ny = lin(x)\nprint('\\n\\nOutput:')\nprint(y)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\u5982\u679c\u4f60\u5c06 ``x`` \u4e0e\u7ebf\u6027\u5c42\u7684\u6743\u91cd\u76f8\u4e58,\u5e76\u52a0\u4e0a\u504f\u7f6e,\u4f60\u4f1a\u53d1\u73b0\u5f97\u5230\u7684\u662f\u8f93\u51fa\u5411\u91cf ``y``\u3002\n\n\u53e6\u4e00\u4e2a\u9700\u8981\u6ce8\u610f\u7684\u91cd\u8981\u7279\u6027\u662f:\u5f53\u6211\u4eec\u7528 ``lin.weight`` \u68c0\u67e5\u5c42\u7684\u6743\u91cd\u65f6,\n\u5b83\u5c06\u81ea\u5df1\u62a5\u544a\u4e3a\u4e00\u4e2a ``Parameter``(\u8fd9\u662f ``Tensor`` \u7684\u5b50\u7c7b),\n\u5e76\u8ba9\u6211\u4eec\u77e5\u9053\u5b83\u6b63\u5728\u4f7f\u7528 autograd \u8ddf\u8e2a\u68af\u5ea6\u3002\u8fd9\u662f ``Parameter`` \u4e0e ``Tensor`` \u4e0d\u540c\u7684\u9ed8\u8ba4\u884c\u4e3a\u3002\n\n\u7ebf\u6027\u5c42\u5728\u6df1\u5ea6\u5b66\u4e60\u6a21\u578b\u4e2d\u88ab\u5e7f\u6cdb\u4f7f\u7528\u3002\u4f60\u4f1a\u7ecf\u5e38\u5728\u5206\u7c7b\u5668\u6a21\u578b\u7684\u672b\u7aef\u770b\u5230\u5b83\u4eec,\u5176\u4e2d\u6700\u540e\u4e00\u5c42\u5c06\u6709 *n* \u4e2a\u8f93\u51fa,\n\u5176\u4e2d *n* \u662f\u5206\u7c7b\u5668\u6240\u5904\u7406\u7684\u7c7b\u522b\u6570\u3002\n\n### \u5377\u79ef\u5c42\n\n*\u5377\u79ef* \u5c42\u88ab\u8bbe\u8ba1\u7528\u4e8e\u5904\u7406\u5177\u6709\u9ad8\u5ea6\u7a7a\u95f4\u76f8\u5173\u6027\u7684\u6570\u636e\u3002\u5b83\u4eec\u5728\u8ba1\u7b97\u673a\u89c6\u89c9\u9886\u57df\u975e\u5e38\u5e38\u7528,\n\u7528\u4e8e\u68c0\u6d4b\u7ec4\u6210\u66f4\u9ad8\u7ea7\u7279\u5f81\u7684\u7d27\u5bc6\u7279\u5f81\u7ec4\u3002\u5b83\u4eec\u4e5f\u51fa\u73b0\u5728\u5176\u4ed6\u4e0a\u4e0b\u6587\u4e2d - \u4f8b\u5982,\u5728 NLP \u5e94\u7528\u7a0b\u5e8f\u4e2d,\n\u4e00\u4e2a\u5355\u8bcd\u7684\u76f4\u63a5\u4e0a\u4e0b\u6587(\u5373\u5e8f\u5217\u4e2d\u9644\u8fd1\u7684\u5176\u4ed6\u5355\u8bcd)\u53ef\u80fd\u4f1a\u5f71\u54cd\u53e5\u5b50\u7684\u542b\u4e49\u3002\n\n\u6211\u4eec\u5728\u4e4b\u524d\u7684\u89c6\u9891\u4e2d\u770b\u5230\u4e86 LeNet5 \u4e2d\u7684\u5377\u79ef\u5c42:\n\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import torch.functional as F\n\n\nclass LeNet(torch.nn.Module):\n\n    def __init__(self):\n        super(LeNet, self).__init__()\n        # 1 \u8f93\u5165\u56fe\u50cf\u901a\u9053(\u9ed1\u767d), 6 \u8f93\u51fa\u901a\u9053, 5x5 \u5e73\u65b9\u5377\u79ef\u6838\n        self.conv1 = torch.nn.Conv2d(1, 6, 5)\n        self.conv2 = torch.nn.Conv2d(6, 16, 3)\n        # \u4e00\u4e2a\u4eff\u5c04\u64cd\u4f5c: y = Wx + b\n        self.fc1 = torch.nn.Linear(16 * 6 * 6, 120)  # 6*6 \u6765\u81ea\u56fe\u50cf\u7ef4\u5ea6\n        self.fc2 = torch.nn.Linear(120, 84)\n        self.fc3 = torch.nn.Linear(84, 10)\n\n    def forward(self, x):\n        # \u5728 (2, 2) \u7a97\u53e3\u4e0a\u8fdb\u884c\u6700\u5927\u6c60\u5316\n        x = F.max_pool2d(F.relu(self.conv1(x)), (2, 2))\n        # \u5982\u679c\u5c3a\u5bf8\u662f\u6b63\u65b9\u5f62,\u4f60\u53ea\u9700\u6307\u5b9a\u4e00\u4e2a\u6570\u5b57\n        x = F.max_pool2d(F.relu(self.conv2(x)), 2)\n        x = x.view(-1, self.num_flat_features(x))\n        x = F.relu(self.fc1(x))\n        x = F.relu(self.fc2(x))\n        x = self.fc3(x)\n        return x\n\n    def num_flat_features(self, x):\n        size = x.size()[1:]  # \u9664\u6279\u6b21\u7ef4\u5ea6\u5916\u7684\u6240\u6709\u7ef4\u5ea6\n        num_features = 1\n        for s in size:\n            num_features *= s\n        return num_features"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\u8ba9\u6211\u4eec\u5206\u89e3\u4e00\u4e0b\u8fd9\u4e2a\u6a21\u578b\u4e2d\u5377\u79ef\u5c42\u7684\u5de5\u4f5c\u539f\u7406\u3002\u4ece ``conv1`` \u5f00\u59cb:\n\n-  LeNet5 \u65e8\u5728\u63a5\u53d7 1x32x32 \u7684\u9ed1\u767d\u56fe\u50cf\u3002**\u5377\u79ef\u5c42\u6784\u9020\u51fd\u6570\u7684\u7b2c\u4e00\u4e2a\u53c2\u6570\u662f\u8f93\u5165\u901a\u9053\u6570\u3002**\u8fd9\u91cc\u662f 1\u3002\n   \u5982\u679c\u6211\u4eec\u6784\u5efa\u8fd9\u4e2a\u6a21\u578b\u6765\u67e5\u770b 3 \u8272\u5f69\u901a\u9053,\u5b83\u5c06\u662f 3\u3002\n-  \u5377\u79ef\u5c42\u5c31\u50cf\u4e00\u4e2a\u626b\u63cf\u56fe\u50cf\u7684\u7a97\u53e3,\u5bfb\u627e\u5b83\u80fd\u8bc6\u522b\u7684\u6a21\u5f0f\u3002\u8fd9\u4e9b\u6a21\u5f0f\u88ab\u79f0\u4e3a *\u7279\u5f81*,\n   \u5377\u79ef\u5c42\u7684\u4e00\u4e2a\u53c2\u6570\u662f\u6211\u4eec\u5e0c\u671b\u5b83\u5b66\u4e60\u7684\u7279\u5f81\u6570\u91cf\u3002**\u6784\u9020\u51fd\u6570\u7684\u7b2c\u4e8c\u4e2a\u53c2\u6570\u662f\u8f93\u51fa\u7279\u5f81\u7684\u6570\u91cf\u3002** \n   \u8fd9\u91cc,\u6211\u4eec\u8981\u6c42\u6211\u4eec\u7684\u5c42\u5b66\u4e60 6 \u4e2a\u7279\u5f81\u3002\n-  \u5c31\u5728\u4e0a\u9762,\u6211\u5c06\u5377\u79ef\u5c42\u6bd4\u4f5c\u4e00\u4e2a\u7a97\u53e3 - \u4f46\u662f\u7a97\u53e3\u6709\u591a\u5927? **\u7b2c\u4e09\u4e2a\u53c2\u6570\u662f\u7a97\u53e3\u6216\u5185\u6838\u5927\u5c0f\u3002** \n   \u8fd9\u91cc,\u6570\u5b57 \"5\" \u610f\u5473\u7740\u6211\u4eec\u9009\u62e9\u4e86\u4e00\u4e2a 5x5 \u7684\u5185\u6838\u3002\n   (\u5982\u679c\u4f60\u5e0c\u671b\u5185\u6838\u7684\u9ad8\u5ea6\u4e0e\u5bbd\u5ea6\u4e0d\u540c,\u4f60\u53ef\u4ee5\u4e3a\u6b64\u53c2\u6570\u6307\u5b9a\u4e00\u4e2a\u5143\u7ec4 - \u4f8b\u5982 ``(3, 5)`` \u6765\u83b7\u5f97\u4e00\u4e2a 3x5 \u7684\u5377\u79ef\u6838\u3002)\n\n\u5377\u79ef\u5c42\u7684\u8f93\u51fa\u662f\u4e00\u4e2a *\u6fc0\u6d3b\u6620\u5c04* - \u8f93\u5165\u5f20\u91cf\u4e2d\u7279\u5f81\u5b58\u5728\u7684\u7a7a\u95f4\u8868\u793a\u3002``conv1`` \u5c06\u7ed9\u6211\u4eec\u4e00\u4e2a 6x28x28 \u7684\u8f93\u51fa\u5f20\u91cf;\n6 \u662f\u7279\u5f81\u6570,28 \u662f\u6620\u5c04\u7684\u9ad8\u5ea6\u548c\u5bbd\u5ea6\u3002(28 \u6765\u81ea\u4e8e\u5f53\u5728 32 \u50cf\u7d20\u884c\u4e0a\u626b\u63cf 5 \u50cf\u7d20\u7a97\u53e3\u65f6,\u53ea\u6709 28 \u4e2a\u6709\u6548\u4f4d\u7f6e\u7684\u4e8b\u5b9e\u3002)\n\n\u7136\u540e,\u6211\u4eec\u5c06\u5377\u79ef\u7684\u8f93\u51fa\u901a\u8fc7 ReLU \u6fc0\u6d3b\u51fd\u6570(\u7a0d\u540e\u5c06\u8ba8\u8bba\u6fc0\u6d3b\u51fd\u6570),\u7136\u540e\u901a\u8fc7\u6700\u5927\u6c60\u5316\u5c42\u3002\n\u6700\u5927\u6c60\u5316\u5c42\u5c06\u6fc0\u6d3b\u6620\u5c04\u4e2d\u5f7c\u6b64\u9760\u8fd1\u7684\u7279\u5f81\u7ec4\u5408\u5728\u4e00\u8d77\u3002\u5b83\u901a\u8fc7\u51cf\u5c0f\u5f20\u91cf\u6765\u5b9e\u73b0\u8fd9\u4e00\u70b9,\n\u5c06\u8f93\u51fa\u4e2d\u6bcf\u4e2a 2x2 \u7ec4\u7684\u5355\u5143\u683c\u5408\u5e76\u4e3a\u4e00\u4e2a\u5355\u5143\u683c,\u5e76\u5c06\u8be5\u5355\u5143\u683c\u7684\u503c\u5206\u914d\u4e3a\u7ec4\u6210\u5b83\u7684 4 \u4e2a\u5355\u5143\u683c\u4e2d\u7684\u6700\u5927\u503c\u3002\n\u8fd9\u7ed9\u4e86\u6211\u4eec\u4e00\u4e2a\u8f83\u4f4e\u5206\u8fa8\u7387\u7684\u6fc0\u6d3b\u6620\u5c04,\u5c3a\u5bf8\u4e3a 6x14x14\u3002\n\n\u6211\u4eec\u7684\u4e0b\u4e00\u4e2a\u5377\u79ef\u5c42 ``conv2`` \u671f\u671b 6 \u4e2a\u8f93\u5165\u901a\u9053(\u5bf9\u5e94\u4e8e\u7b2c\u4e00\u5c42\u5bfb\u627e\u7684 6 \u4e2a\u7279\u5f81),\u6709 16 \u4e2a\u8f93\u51fa\u901a\u9053,\n\u5e76\u4e14\u5185\u6838\u5927\u5c0f\u4e3a 3x3\u3002\u5b83\u8f93\u51fa\u4e00\u4e2a 16x12x12 \u7684\u6fc0\u6d3b\u6620\u5c04,\u518d\u6b21\u901a\u8fc7\u6700\u5927\u6c60\u5316\u5c42\u51cf\u5c0f\u5230 16x6x6\u3002\n\u5728\u5c06\u6b64\u8f93\u51fa\u4f20\u9012\u7ed9\u7ebf\u6027\u5c42\u4e4b\u524d,\u5b83\u88ab\u91cd\u65b0\u6574\u5f62\u4e3a\u4e00\u4e2a 16 * 6 * 6 = 576 \u5143\u7d20\u5411\u91cf,\u4f9b\u4e0b\u4e00\u5c42\u4f7f\u7528\u3002\n\n\u6709\u9488\u5bf9 1D\u30012D \u548c 3D \u5f20\u91cf\u7684\u5377\u79ef\u5c42\u3002\u5377\u79ef\u5c42\u6784\u9020\u51fd\u6570\u8fd8\u6709\u8bb8\u591a\u5176\u4ed6\u53ef\u9009\u53c2\u6570,\n\u5305\u62ec\u6b65\u957f\u957f\u5ea6(\u4f8b\u5982,\u53ea\u626b\u63cf\u6bcf\u7b2c\u4e8c\u4e2a\u6216\u6bcf\u7b2c\u4e09\u4e2a\u4f4d\u7f6e)\u3001\u586b\u5145(\u56e0\u6b64\u4f60\u53ef\u4ee5\u626b\u63cf\u5230\u8f93\u5165\u7684\u8fb9\u7f18)\u7b49\u7b49\u3002\n\u66f4\u591a\u4fe1\u606f\u8bf7\u53c2\u89c1 [\u6587\u6863](https://pytorch.org/docs/stable/nn.html#convolution-layers)_\u3002\n\n### \u5faa\u73af\u5c42\n\n*\u5faa\u73af\u795e\u7ecf\u7f51\u7edc* (\u6216 *RNN*)\u7528\u4e8e\u5e8f\u5217\u6570\u636e - \u4ece\u79d1\u5b66\u4eea\u5668\u7684\u65f6\u95f4\u5e8f\u5217\u6d4b\u91cf\u5230\u81ea\u7136\u8bed\u8a00\u53e5\u5b50\u518d\u5230 DNA \u6838\u82f7\u9178\u3002\nRNN \u901a\u8fc7\u7ef4\u62a4\u4e00\u4e2a *\u9690\u85cf\u72b6\u6001* \u6765\u5b9e\u73b0\u8fd9\u4e00\u70b9,\u8be5\u9690\u85cf\u72b6\u6001\u5145\u5f53\u4e00\u79cd\u8bb0\u5fc6,\u8bb0\u5f55\u5230\u76ee\u524d\u4e3a\u6b62\u5b83\u5728\u5e8f\u5217\u4e2d\u770b\u5230\u7684\u5185\u5bb9\u3002\n\nRNN \u5c42\u7684\u5185\u90e8\u7ed3\u6784 - \u6216\u5176\u53d8\u4f53 LSTM(\u957f\u77ed\u671f\u8bb0\u5fc6)\u548c GRU(\u95e8\u63a7\u5faa\u73af\u5355\u5143) - \u76f8\u5f53\u590d\u6742,\u8d85\u51fa\u4e86\u672c\u89c6\u9891\u7684\u8303\u56f4,\n\u4f46\u6211\u4eec\u5c06\u5411\u4f60\u5c55\u793a\u57fa\u4e8e LSTM \u7684\u8bcd\u6027\u6807\u6ce8\u5668(\u4e00\u79cd\u5206\u7c7b\u5668,\u7528\u4e8e\u544a\u8bc9\u4f60\u4e00\u4e2a\u5355\u8bcd\u662f\u540d\u8bcd\u3001\u52a8\u8bcd\u7b49)\u7684\u6837\u5b50:\n\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "class LSTMTagger(torch.nn.Module):\n\n    def __init__(self, embedding_dim, hidden_dim, vocab_size, tagset_size):\n        super(LSTMTagger, self).__init__()\n        self.hidden_dim = hidden_dim\n\n        self.word_embeddings = torch.nn.Embedding(vocab_size, embedding_dim)\n\n        # LSTM \u63a5\u53d7\u8bcd\u5d4c\u5165\u4f5c\u4e3a\u8f93\u5165,\u5e76\u8f93\u51fa\u7ef4\u5ea6\u4e3a hidden_dim \u7684\u9690\u85cf\u72b6\u6001\n        self.lstm = torch.nn.LSTM(embedding_dim, hidden_dim)\n\n        # \u5c06\u4ece\u9690\u85cf\u72b6\u6001\u7a7a\u95f4\u6620\u5c04\u5230\u6807\u8bb0\u7a7a\u95f4\u7684\u7ebf\u6027\u5c42\n        self.hidden2tag = torch.nn.Linear(hidden_dim, tagset_size)\n\n    def forward(self, sentence):\n        embeds = self.word_embeddings(sentence)\n        lstm_out, _ = self.lstm(embeds.view(len(sentence), 1, -1))\n        tag_space = self.hidden2tag(lstm_out.view(len(sentence), -1))\n        tag_scores = F.log_softmax(tag_space, dim=1)\n        return tag_scores"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\u6784\u9020\u51fd\u6570\u6709\u56db\u4e2a\u53c2\u6570:\n\n- ``vocab_size`` \u662f\u8f93\u5165\u8bcd\u6c47\u8868\u4e2d\u5355\u8bcd\u7684\u6570\u91cf\u3002\u6bcf\u4e2a\u5355\u8bcd\u662f\u4e00\u4e2a ``vocab_size`` \u7ef4\u7684\u4e00\u70ed\u5411\u91cf(\u6216\u5355\u4f4d\u5411\u91cf)\u3002\n\n- ``tagset_size`` \u662f\u8f93\u51fa\u6807\u7b7e\u96c6\u7684\u5927\u5c0f\u3002\n\n- ``embedding_dim`` \u662f\u8bcd\u6c47\u7684*\u5d4c\u5165*\u7a7a\u95f4\u7684\u5927\u5c0f\u3002\u5d4c\u5165\u5c06\u8bcd\u6c47\u6620\u5c04\u5230\u4e00\u4e2a\u4f4e\u7ef4\u7a7a\u95f4,\u5728\u8be5\u7a7a\u95f4\u4e2d,\u610f\u4e49\u76f8\u4f3c\u7684\u5355\u8bcd\u5f7c\u6b64\u63a5\u8fd1\u3002\n\n- ``hidden_dim`` \u662f LSTM \u7684\u8bb0\u5fc6\u5927\u5c0f\u3002\n\n\u8f93\u5165\u5c06\u662f\u4e00\u4e2a\u53e5\u5b50,\u5355\u8bcd\u8868\u793a\u4e3a\u4e00\u70ed\u5411\u91cf\u7684\u7d22\u5f15\u3002\u5d4c\u5165\u5c42\u5c06\u628a\u8fd9\u4e9b\u6620\u5c04\u5230\u4e00\u4e2a ``embedding_dim`` \u7ef4\u7684\u7a7a\u95f4\u3002\nLSTM \u63a5\u6536\u8fd9\u4e2a\u5d4c\u5165\u5e8f\u5217\u5e76\u5bf9\u5176\u8fdb\u884c\u8fed\u4ee3,\u4ea7\u751f\u4e00\u4e2a\u957f\u5ea6\u4e3a ``hidden_dim`` \u7684\u8f93\u51fa\u5411\u91cf\u3002\u6700\u540e\u7684\u7ebf\u6027\u5c42\u5145\u5f53\u5206\u7c7b\u5668;\n\u5c06\u6700\u540e\u4e00\u5c42\u7684\u8f93\u51fa\u901a\u8fc7 ``log_softmax()`` \u8f6c\u6362\u4e3a\u4e00\u7ec4\u5f52\u4e00\u5316\u7684\u4f30\u8ba1\u6982\u7387,\u8868\u793a\u7ed9\u5b9a\u5355\u8bcd\u6620\u5c04\u5230\u7ed9\u5b9a\u6807\u7b7e\u7684\u6982\u7387\u3002\n\n\u5982\u679c\u4f60\u60f3\u770b\u770b\u8fd9\u4e2a\u7f51\u7edc\u7684\u5b9e\u9645\u8fd0\u884c\u60c5\u51b5,\u53ef\u4ee5\u67e5\u770b pytorch.org \u4e0a\u7684 [\u5e8f\u5217\u6a21\u578b\u548c \nLSTM \u7f51\u7edc](https://pytorch.org/tutorials/beginner/nlp/sequence_models_tutorial.html)_ \u6559\u7a0b\u3002\n\n### \u8f6c\u6362\u5668\n\n*\u8f6c\u6362\u5668*\u662f\u591a\u7528\u9014\u7f51\u7edc,\u5728 NLP \u9886\u57df\u53d6\u5f97\u4e86\u6700\u5148\u8fdb\u7684\u6210\u679c,\u4f8b\u5982 BERT \u6a21\u578b\u3002\u8ba8\u8bba\u8f6c\u6362\u5668\u67b6\u6784\u8d85\u51fa\u4e86\u672c\u89c6\u9891\u7684\u8303\u56f4,\n\u4f46\u662f PyTorch \u6709\u4e00\u4e2a ``Transformer`` \u7c7b,\u5141\u8bb8\u4f60\u5b9a\u4e49\u8f6c\u6362\u5668\u6a21\u578b\u7684\u6574\u4f53\u53c2\u6570 - \u6ce8\u610f\u529b\u5934\u7684\u6570\u91cf\u3001\n\u7f16\u7801\u5668\u548c\u89e3\u7801\u5668\u5c42\u7684\u6570\u91cf\u3001dropout \u548c\u6fc0\u6d3b\u51fd\u6570\u7b49\u3002(\u4f60\u751a\u81f3\u53ef\u4ee5\u7528\u6b63\u786e\u7684\u53c2\u6570\u4ece\u8fd9\u4e2a\u5355\u4e00\u7c7b\u6784\u5efa BERT \u6a21\u578b!)\n ``torch.nn.Transformer`` \u7c7b\u8fd8\u5305\u542b\u5c01\u88c5\u5355\u4e2a\u7ec4\u4ef6(``TransformerEncoder``\u3001``TransformerDecoder``)\n\u548c\u5b50\u7ec4\u4ef6(``TransformerEncoderLayer``\u3001``TransformerDecoderLayer``)\u7684\u7c7b\u3002\n\u8be6\u60c5\u8bf7\u67e5\u770b [\u6587\u6863](https://pytorch.org/docs/stable/nn.html#transformer-layers)_ \u4e2d\u5173\u4e8e\u8f6c\u6362\u5668\u7c7b\u7684\u5185\u5bb9,\n\u4ee5\u53ca pytorch.org \u4e0a\u76f8\u5173\u7684 [\u6559\u7a0b](https://pytorch.org/tutorials/beginner/transformer_tutorial.html)_\u3002\n\n## \u5176\u4ed6\u5c42\u548c\u51fd\u6570\n\n### \u6570\u636e\u64cd\u4f5c\u5c42\n\n\u8fd8\u6709\u5176\u4ed6\u7c7b\u578b\u7684\u5c42\u6267\u884c\u6a21\u578b\u4e2d\u7684\u91cd\u8981\u529f\u80fd,\u4f46\u5b83\u4eec\u81ea\u8eab\u4e0d\u53c2\u4e0e\u5b66\u4e60\u8fc7\u7a0b\u3002\n\n**\u6700\u5927\u6c60\u5316**(\u53ca\u5176\u5b6a\u751f\u5c42\u6700\u5c0f\u6c60\u5316)\u901a\u8fc7\u7ec4\u5408\u5355\u5143\u5e76\u5c06\u8f93\u5165\u5355\u5143\u7684\u6700\u5927\u503c\u5206\u914d\u7ed9\u8f93\u51fa\u5355\u5143\u6765\u51cf\u5c0f\u5f20\u91cf(\u6211\u4eec\u4e4b\u524d\u770b\u5230\u8fc7\u8fd9\u4e00\u70b9)\u3002\n\u4f8b\u5982:\n\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "my_tensor = torch.rand(1, 6, 6)\nprint(my_tensor)\n\nmaxpool_layer = torch.nn.MaxPool2d(3)\nprint(maxpool_layer(my_tensor))"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\u5982\u679c\u4f60\u4ed4\u7ec6\u89c2\u5bdf\u4e0a\u9762\u7684\u503c,\u4f60\u4f1a\u53d1\u73b0\u6700\u5927\u6c60\u5316\u8f93\u51fa\u4e2d\u7684\u6bcf\u4e2a\u503c\u90fd\u662f 6x6 \u8f93\u5165\u7684\u6bcf\u4e2a\u8c61\u9650\u7684\u6700\u5927\u503c\u3002\n\n**\u5f52\u4e00\u5316\u5c42**\u5728\u5c06\u4e00\u5c42\u7684\u8f93\u51fa\u9988\u9001\u5230\u53e6\u4e00\u5c42\u4e4b\u524d,\u91cd\u65b0\u5c45\u4e2d\u5e76\u5f52\u4e00\u5316\u8f93\u51fa\u3002\u5c45\u4e2d\u548c\u7f29\u653e\u4e2d\u95f4\u5f20\u91cf\u6709\u8bb8\u591a\u6709\u76ca\u7684\u6548\u679c,\n\u4f8b\u5982\u8ba9\u4f60\u53ef\u4ee5\u4f7f\u7528\u66f4\u9ad8\u7684\u5b66\u4e60\u7387\u800c\u4e0d\u4f1a\u51fa\u73b0\u68af\u5ea6\u7206\u70b8/\u6d88\u5931\u3002\n\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "my_tensor = torch.rand(1, 4, 4) * 20 + 5\nprint(my_tensor)\n\nprint(my_tensor.mean())\n\nnorm_layer = torch.nn.BatchNorm1d(4)\nnormed_tensor = norm_layer(my_tensor)\nprint(normed_tensor)\n\nprint(normed_tensor.mean())"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\u8fd0\u884c\u4e0a\u9762\u7684\u5355\u5143\u683c,\u6211\u4eec\u4e3a\u8f93\u5165\u5f20\u91cf\u6dfb\u52a0\u4e86\u4e00\u4e2a\u5927\u7684\u7f29\u653e\u56e0\u5b50\u548c\u504f\u79fb\u91cf;\u4f60\u5e94\u8be5\u4f1a\u770b\u5230\u8f93\u5165\u5f20\u91cf\u7684 ``mean()`` \u5728 15 \u5de6\u53f3\u3002\n\u7ecf\u8fc7\u5f52\u4e00\u5316\u5c42\u5904\u7406\u540e,\u4f60\u53ef\u4ee5\u770b\u5230\u503c\u53d8\u5c0f\u4e86,\u5e76\u4e14\u56f4\u7ed5\u7740 0 \u5206\u5e03 - \u4e8b\u5b9e\u4e0a,\u5e73\u5747\u503c\u5e94\u8be5\u975e\u5e38\u5c0f(> 1e-8)\u3002\n\n\u8fd9\u662f\u6709\u76ca\u7684,\u56e0\u4e3a\u8bb8\u591a\u6fc0\u6d3b\u51fd\u6570(\u4e0b\u9762\u5c06\u8ba8\u8bba)\u5728 0 \u9644\u8fd1\u5177\u6709\u6700\u5f3a\u68af\u5ea6,\u4f46\u6709\u65f6\u5bf9\u4e8e\u5c06\u5b83\u4eec\u63a8\u79bb 0 \u5f88\u8fdc\u7684\u8f93\u5165\u4f1a\u9047\u5230\u68af\u5ea6\u6d88\u5931\u6216\u7206\u70b8\u7684\u95ee\u9898\u3002\n\u5c06\u6570\u636e\u4fdd\u6301\u5728\u6700\u9661\u68af\u5ea6\u533a\u57df\u5468\u56f4\u5c06\u503e\u5411\u4e8e\u610f\u5473\u7740\u66f4\u5feb\u3001\u66f4\u597d\u7684\u5b66\u4e60\u548c\u66f4\u9ad8\u7684\u53ef\u884c\u5b66\u4e60\u7387\u3002\n\n**Dropout \u5c42**\u662f\u4e00\u79cd\u9f13\u52b1\u6a21\u578b*\u7a00\u758f\u8868\u793a*\u7684\u5de5\u5177 - \u4e5f\u5c31\u662f\u8bf4,\u63a8\u52a8\u5b83\u5728\u63a8\u7406\u65f6\u4f7f\u7528\u8f83\u5c11\u7684\u6570\u636e\u3002\n\nDropout \u5c42\u7684\u5de5\u4f5c\u539f\u7406\u662f\u5728*\u8bad\u7ec3\u671f\u95f4*\u968f\u673a\u8bbe\u7f6e\u8f93\u5165\u5f20\u91cf\u7684\u4e00\u90e8\u5206 - dropout \u5c42\u5728\u63a8\u7406\u65f6\u603b\u662f\u5173\u95ed\u7684\u3002\n\u8fd9\u8feb\u4f7f\u6a21\u578b\u9488\u5bf9\u8fd9\u79cd\u63a9\u7801\u6216\u51cf\u5c11\u7684\u6570\u636e\u96c6\u8fdb\u884c\u5b66\u4e60\u3002\u4f8b\u5982:\n\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "my_tensor = torch.rand(1, 4, 4)\n\ndropout = torch.nn.Dropout(p=0.4)\nprint(dropout(my_tensor))\nprint(dropout(my_tensor))"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\u4e0a\u9762,\u4f60\u53ef\u4ee5\u770b\u5230 dropout \u5bf9\u793a\u4f8b\u5f20\u91cf\u7684\u5f71\u54cd\u3002\u4f60\u53ef\u4ee5\u4f7f\u7528\u53ef\u9009\u7684 ``p`` \u53c2\u6570\u8bbe\u7f6e\u5355\u4e2a\u6743\u91cd\u4e22\u5f03\u7684\u6982\u7387;\n\u5982\u679c\u4e0d\u8bbe\u7f6e,\u9ed8\u8ba4\u4e3a 0.5\u3002\n\n### \u6fc0\u6d3b\u51fd\u6570\n\n\u6fc0\u6d3b\u51fd\u6570\u4f7f\u6df1\u5ea6\u5b66\u4e60\u6210\u4e3a\u53ef\u80fd\u3002\u795e\u7ecf\u7f51\u7edc\u5b9e\u9645\u4e0a\u662f\u4e00\u4e2a\u7a0b\u5e8f - \u6709\u8bb8\u591a\u53c2\u6570 - \u7528\u4e8e*\u6a21\u62df\u4e00\u4e2a\u6570\u5b66\u51fd\u6570*\u3002\n\u5982\u679c\u6211\u4eec\u53ea\u662f\u91cd\u590d\u5730\u5c06\u5f20\u91cf\u4e0e\u5c42\u6743\u91cd\u76f8\u4e58,\u6211\u4eec\u53ea\u80fd\u6a21\u62df*\u7ebf\u6027\u51fd\u6570*;\u6b64\u5916,\u6709\u591a\u5c42\u4e5f\u6ca1\u6709\u610f\u4e49,\n\u56e0\u4e3a\u6574\u4e2a\u7f51\u7edc\u53ef\u4ee5\u7b80\u5316\u4e3a\u5355\u4e2a\u77e9\u9635\u4e58\u6cd5\u3002\u5728\u5c42\u4e4b\u95f4\u63d2\u5165*\u975e\u7ebf\u6027*\u6fc0\u6d3b\u51fd\u6570\u4f7f\u5f97\u6df1\u5ea6\u5b66\u4e60\u6a21\u578b\u80fd\u591f\u6a21\u62df\u4efb\u4f55\u51fd\u6570,\n\u800c\u4e0d\u4ec5\u4ec5\u662f\u7ebf\u6027\u51fd\u6570\u3002\n\n``torch.nn.Module`` \u6709\u5c01\u88c5\u6240\u6709\u4e3b\u8981\u6fc0\u6d3b\u51fd\u6570\u7684\u5bf9\u8c61,\u5305\u62ec ReLU \u53ca\u5176\u8bb8\u591a\u53d8\u4f53\u3001Tanh\u3001Hardtanh\u3001sigmoid \u7b49\u3002\n\u5b83\u8fd8\u5305\u62ec\u5176\u4ed6\u51fd\u6570,\u5982 Softmax,\u8fd9\u4e9b\u51fd\u6570\u5728\u6a21\u578b\u7684\u8f93\u51fa\u9636\u6bb5\u6700\u6709\u7528\u3002\n\n### \u635f\u5931\u51fd\u6570\n\n\u635f\u5931\u51fd\u6570\u544a\u8bc9\u6211\u4eec\u6a21\u578b\u7684\u9884\u6d4b\u4e0e\u6b63\u786e\u7b54\u6848\u76f8\u5dee\u591a\u8fdc\u3002PyTorch \u5305\u542b\u5404\u79cd\u635f\u5931\u51fd\u6570,\u5305\u62ec\u5e38\u89c1\u7684 MSE(\u5747\u65b9\u8bef\u5dee = L2 \u8303\u6570)\u3001\n\u4ea4\u53c9\u71b5\u635f\u5931\u548c\u8d1f\u5bf9\u6570\u4f3c\u7136\u635f\u5931(\u5bf9\u4e8e\u5206\u7c7b\u5668\u5f88\u6709\u7528)\u7b49\u3002\n"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.11.9"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}